# Copyright 2025 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Sequence
from typing import Any, overload

from lightworks import Batch, Sampler
from lightworks import __version__ as LIGHTWORKS_VERSION  # noqa: N812
from lightworks.sdk.tasks import Task
from multimethod import multimethod
from numpy import inf

from lightworks_remote.http.qpu_details import QPUDetails
from lightworks_remote.payloads import ArtemisPayload, Payload
from lightworks_remote.utils import ValidationError

from .batch_job import BatchJob
from .job import Job
from .sampler_compiler import SamplerCompiler
from .system_management import get_qpu_details


class QPU:
    """
    Defines the target QPU and facilitates execution of a task on that
    particular piece of hardware.

    Args:

        name (str) : The id of the QPU which is to be used.

    """

    def __init__(self, name: str) -> None:
        # Assign name to attribute
        self.name = name

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f"lightworks_remote.QPU('{self.name}')"

    @property
    def name(self) -> str:
        """Returns the name of the QPU"""
        return self.__name

    @name.setter
    def name(self, value: str) -> None:
        if not isinstance(value, str):
            raise TypeError("QPU name should be a string.")
        self.__name = value

    @property
    def details(self) -> QPUDetails:
        """Returns all details about the selected QPU"""
        return get_qpu_details(self.name)

    @overload
    def __call__(
        self,
        task: Task,
        direct_encoding: bool = ...,
        job_name: str = ...,
        options: dict[str, Any] | None = None,
    ) -> Job: ...

    @overload
    def __call__(
        self,
        task: Batch,
        direct_encoding: bool = ...,
        job_name: str | list[str] = ...,
        options: dict[str, Any] | None = None,
    ) -> BatchJob: ...

    def __call__(
        self,
        task: Task | Batch,
        direct_encoding: bool = False,
        job_name: str | list[str] = "Job",
        options: dict[str, Any] | None = None,
    ) -> Job | BatchJob:
        """Prepare and submit a task to a target QPU backend."""
        return self.run(task, direct_encoding, job_name, options)  # type: ignore[arg-type]

    @overload
    def run(
        self,
        task: Task,
        direct_encoding: bool = ...,
        job_name: str = ...,
        options: dict[str, Any] | None = None,
    ) -> Job: ...

    @overload
    def run(
        self,
        task: Batch,
        direct_encoding: bool = ...,
        job_name: str | list[str] = ...,
        options: dict[str, Any] | None = None,
    ) -> BatchJob: ...

    def run(
        self,
        task: Task | Batch,
        direct_encoding: bool = False,
        job_name: str | list[str] = "Job",
        options: dict[str, Any] | None = None,
    ) -> Job | BatchJob:
        """
        Prepare and submit a task to a target QPU backend.

        Args:

            task (Task | Batch) : The target task to run on the chosen QPU.

            direct_encoding (bool, optional) : Controls whether a circuit is
                compiled into a unitary to be executed or the circuit spec is
                used to better preserve the structure of the circuit.

            job_name (str | list, optional) : Set a name to use as a reference
                for the job. If a batch then this can be a list of names,
                otherwise a number will be appended to the base name for each
                job in the batch. Defaults to job.

            options (dict | None, optional) : Allows additional options to be
                provided within job data. Should only be used by advanced users
                who have been told which options are available.

        Returns:

            Job : Job object which can be used for interacting with the created
                job.

        """
        # Perform initial submission
        job = self.run_async(
            task,  # type: ignore[arg-type]
            direct_encoding=direct_encoding,
            job_name=job_name,  # type: ignore[arg-type]
            options=options,
        )
        # Wait until compilation is complete
        # For single jobs
        if isinstance(job, Job):
            job._check_successful_compilation()
        # For multiple
        else:
            for j in job.jobs:
                j._check_successful_compilation()

        # Then return created job object
        return job

    @overload
    def run_async(
        self,
        task: Task,
        direct_encoding: bool = ...,
        job_name: str = ...,
        options: dict[str, Any] | None = None,
    ) -> Job: ...

    @overload
    def run_async(
        self,
        task: Batch,
        direct_encoding: bool = ...,
        job_name: str | list[str] = ...,
        options: dict[str, Any] | None = None,
    ) -> BatchJob: ...

    def run_async(
        self,
        task: Task | Batch,
        direct_encoding: bool = False,
        job_name: str | list[str] = "Job",
        options: dict[str, Any] | None = None,
    ) -> Job | BatchJob:
        """
        Prepare and submit a task to a target QPU backend without waiting for
        compilation to complete.

        Args:

            task (Task | Batch) : The target task to run on the chosen QPU.

            direct_encoding (bool, optional) : Controls whether a circuit is
                compiled into a unitary to be executed or the circuit spec is
                used to better preserve the structure of the circuit.

            job_name (str | list, optional) : Set a name to use as a reference
                for the job. If a batch then this can be a list of names,
                otherwise a number will be appended to the base name for each
                job in the batch. Defaults to job.

            options (dict | None, optional) : Allows additional options to be
                provided within job data. Should only be used by advanced users
                who have been told which options are available.

        Returns:

            Job : Job object which can be used for interacting with the created
                job.

        """
        # Check task
        if not isinstance(task, Task | Batch):
            raise TypeError("Provided task should be a Task or Batch object.")
        # Check direct encoding is boolean
        if not isinstance(direct_encoding, bool):
            raise TypeError(
                "direct_encoding should be set to either True or False."
            )
        # Download QPU data to be used for validation
        self._qpu_data = get_qpu_details(self.name)
        # Then run task
        return self._run(task, direct_encoding, job_name, options)

    @multimethod
    def _run(
        self,
        task: Task,
        direct_encoding: bool,
        job_name: str,
        options: dict[str, Any] | None = None,
    ) -> Job:
        """
        Generates payload from a single task and submits.
        """
        # Process job name
        job_name = get_job_name(task, job_name)
        # Get payload and add QPU + job name
        payload = self._get_payload(task, direct_encoding, job_name, options)

        # And finally submit job and return
        job = Job(payload)
        job._submit()
        return job

    @_run.register
    def _run_batch(
        self,
        task: Batch,
        direct_encoding: bool,
        job_name: str | list[str],
        options: dict[str, Any] | None = None,
    ) -> BatchJob:
        """
        Generates the payload from a batch of tasks and submits these.
        """
        # Process job name
        job_name = get_job_name(task, job_name)
        # Create batch job and add individuals
        batch_job = BatchJob()
        for i, c in enumerate(task):
            # Get payload and add required additions
            payload = self._get_payload(
                c, direct_encoding, job_name[i], options
            )

            # Store job in batch jobs
            batch_job[payload.job_name] = Job(payload)  # type: ignore[attr-defined]
        # Then submit all
        batch_job._submit_all()
        return batch_job

    def _get_payload(
        self,
        task: Task,
        direct_encoding: bool,
        job_name: str,
        options: dict[str, Any] | None = None,
    ) -> Payload:
        """
        Generates a verifies a payload from a task and then returns this.
        """
        payload = self._build_payload_from_task(
            task, direct_encoding, job_name, options
        )
        # Check payload before returning
        self._verify_payload(payload)
        return payload

    def _build_payload_from_task(
        self,
        task: Task,
        direct_encoding: bool,
        job_name: str,
        options: dict[str, Any] | None = None,
    ) -> ArtemisPayload:
        """
        Creates a payload using data from the provided task.
        """
        if not isinstance(task, Sampler):
            raise ValidationError(
                "Remote QPU only supports execution of Sampler tasks."
            )
        data = task._generate_task()
        compiler = SamplerCompiler(data)
        payload = ArtemisPayload(
            qpu=self.name,
            job_name=job_name,
            lightworks_version=LIGHTWORKS_VERSION,
            **compiler._generate_payload(direct_encoding=direct_encoding),
        )
        # Add additional job data from kwargs
        if options is not None:
            for k, v in options.items():
                if payload.job_data is None:
                    payload.job_data = {}
                payload.job_data[k] = v
        return payload

    def _verify_payload(self, payload: ArtemisPayload) -> None:
        """
        Checks the payload against data from the QPU using a set of validation
        rules.
        """
        # Number of modes
        if payload.n_modes > self._qpu_data.n_modes:
            msg = (
                f"Number of modes required for circuit ({payload.n_modes}) is "
                "larger the number the target QPU "
                f"({self._qpu_data.n_modes})."
            )
            raise ValidationError(msg)
        # Input state
        if sum(payload.input) > self._qpu_data.max_photon_input:
            msg = (
                f"Number of input photons ({sum(payload.input)}) is larger "
                "than the maximum available for the system "
                f"({self._qpu_data.max_photon_input})."
            )
            raise ValidationError(msg)
        # Min photon detection number
        if payload.min_detection > self._qpu_data.max_detection_filter:
            msg = (
                f"Minimum photon detection filter ({payload.min_detection}) is "
                "larger than the max value allowed for the system "
                f"({self._qpu_data.max_detection_filter})."
            )
            raise ValidationError(msg)
        # Number of samples
        allowed_samples = min(
            self._qpu_data.default_max_samples,
            self._qpu_data.max_samples.get(payload.min_detection, inf),
        )
        if payload.n_samples > allowed_samples:
            msg = (
                f"Number of samples ({payload.n_samples:,}) is larger "
                "than the maximum allowed for the set minimum photon detection "
                f"filter ({allowed_samples:,})."
            )
            raise ValidationError(msg)


@overload
def get_job_name(task: Task, name: str) -> str: ...


@overload
def get_job_name(task: Batch, name: str | list[str]) -> list[str]: ...


def get_job_name(task: Task | Batch, name: str | list[str]) -> str | list[str]:
    """
    Processes a provided job names, checking it is of the correct format, and
    returns the value.
    """
    if not isinstance(name, str):
        if isinstance(task, Batch) and isinstance(name, Sequence):
            if len(name) == task.num:
                return list(name)
            msg = (
                "Number of job names should be equal to the number of "
                f"tasks ({task.num})."
            )
            raise ValueError(msg)
        if isinstance(task, Batch):
            raise TypeError("Job name should be a string or list of strings.")
        raise TypeError("Job name should be a string.")
    if isinstance(task, Batch):
        return [name + f"_{i + 1}" for i in range(task.num)]
    return name



# Copyright 2025 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import pytest
from lightworks import State

from lightworks_remote.results import QPUSamplingResult
from lightworks_remote.results.qpu_result import (
    apply_post_selection,
    meets_post_selection,
)


class TestResult:
    """
    Unit tests for the QPUSamplingResult object.
    """

    def setup_method(self):
        """
        Creates a set of test data that can be used with the result.
        """
        self.data = {
            "results": {
                State([1, 1, 1, 0, 0, 0]): 100,
                State([0, 0, 0, 1, 1, 1]): 200,
                State([1, 0, 1, 1, 0, 0]): 300,
                State([0, 1, 0, 0, 1, 1]): 400,
            }
        }

    def test_result_creation(self):
        """
        Checks QPUSamplingResult object can be created from a provided data
        dictionary.
        """
        QPUSamplingResult(self.data)

    def test_result_creation_post_selection(self):
        """
        Checks QPUSamplingResult object can be created from a provided data
        dictionary when post-selection is included.
        """
        self.data["post_selection"] = {"post_select": [((0, 1), (1,))]}
        QPUSamplingResult(self.data)

    def test_result_creation_post_selection_herald(self):
        """
        Checks QPUSamplingResult object can be created from a provided data
        dictionary when post-selection is included only through heralding.
        """
        self.data["post_selection"] = {"herald": [((0, 1), (1,))]}
        QPUSamplingResult(self.data)

    @pytest.mark.parametrize("value", [[], None])
    def test_result_creation_post_selection_empty_post_select(self, value):
        """
        Checks QPUSamplingResult object can be created from a provided data
        dictionary when post-selection is included.
        """
        self.data["post_selection"] = {"post_select": value}
        QPUSamplingResult(self.data)

    @pytest.mark.parametrize("value", [[], None])
    def test_result_creation_post_selection_empty_herald(self, value):
        """
        Checks QPUSamplingResult object can be created from a provided data
        dictionary when post-selection is included.
        """
        self.data["post_selection"] = {
            "post_select": [((0, 1), (1,))],
            "herald": value,
        }
        QPUSamplingResult(self.data)

    def test_post_selection_not_in_data(self):
        """
        Checks post-selection field is correctly removed from data and raw data.
        """
        self.data["job_data"] = {"post_select": [((0, 1), (1,))]}
        r = QPUSamplingResult(self.data)
        assert "post_selection" not in r
        assert "post_selection" not in r.raw_data

    def test_post_selection_applied(self):
        """
        Checks that post-selection is correctly applied to remove states which
        do not match.
        """
        self.data["job_data"] = {"post_select": [((0,), (1,))]}
        r = QPUSamplingResult(self.data)
        assert len(r) == 2
        assert State([1, 1, 1, 0, 0, 0]) in r
        assert State([1, 0, 1, 1, 0, 0]) in r
        assert State([0, 0, 0, 1, 1, 1]) not in r
        assert State([0, 1, 0, 0, 1, 1]) not in r

    def test_heralding_applied(self):
        """
        Checks that heralding is correctly applied to remove states which do not
        match.
        """
        self.data["job_data"] = {"herald": [((0,), (1,))]}
        r = QPUSamplingResult(self.data)
        assert len(r) == 2
        assert State([1, 1, 0, 0, 0]) in r
        assert State([0, 1, 1, 0, 0]) in r
        assert State([0, 0, 1, 1, 1]) not in r
        assert State([1, 0, 0, 1, 1]) not in r

    def test_result_does_not_modify_raw_data(self):
        """
        Checks result does not modify raw data when applying processing.
        """
        self.data["job_data"] = {
            "post_select": [((0, 1), (1,))],
            "herald": [((2,), (1,))],
        }
        QPUSamplingResult(self.data)
        assert "job_data" in self.data

    def test_result_raw_data_is_unprocessed(self):
        """
        Checks result does not modify raw data when applying post-selection and
        confirms data and raw data are different.
        """
        self.data["job_data"] = {"post_select": [((0, 1), (1,))]}
        r = QPUSamplingResult(self.data)
        assert r.raw_data == self.data["results"]
        assert r.raw_data != r

    def test_results_save(self):
        """
        Checks result data can be saved to a file.
        """
        r = QPUSamplingResult(self.data)
        r.save("_test_results")
        assert os.path.isfile("_test_results.json")
        os.remove("_test_results.json")

    def test_results_save_json(self):
        """
        Checks result data can be saved to a file when json file extension is
        included at the end.
        """
        r = QPUSamplingResult(self.data)
        r.save("_test_results.json")
        assert os.path.isfile("_test_results.json")
        os.remove("_test_results.json")

    def test_results_save_path(self):
        """
        Checks result data can be saved to a file with a relative path. The
        existing .pytest_cache directory is used.
        """
        r = QPUSamplingResult(self.data)
        r.save(".pytest_cache/_test_results")
        assert os.path.isfile(".pytest_cache/_test_results.json")
        os.remove(".pytest_cache/_test_results.json")

    def test_results_save_abs_path(self):
        """
        Checks result data can be saved to a file with an absolute path. The
        existing .pytest_cache directory is used.
        """
        r = QPUSamplingResult(self.data)
        r.save(os.getcwd() + "/.pytest_cache/_test_results")
        assert os.path.isfile(".pytest_cache/_test_results.json")
        os.remove(".pytest_cache/_test_results.json")

    def test_meets_post_selection(self):
        """
        Checks meets_post_selection returns True for a valid State.
        """
        assert meets_post_selection(State([1, 0, 1, 0]), [((1,), (0,))])
        assert meets_post_selection(State([1, 0, 1, 0]), [((0, 1), (1,))])
        assert meets_post_selection(State([1, 0, 1, 0]), [((0, 1, 2), (2,))])

    def test_meets_post_selection_false(self):
        """
        Checks meets_post_selection returns False for an invalid State.
        """
        assert not meets_post_selection(State([0, 1, 1, 0]), [((1,), (0,))])
        assert not meets_post_selection(State([1, 1, 0, 0]), [((0, 1), (1,))])
        assert not meets_post_selection(
            State([0, 0, 1, 1]), [((0, 1, 2), (2,))]
        )

    def test_meets_post_selection_multi_photons(self):
        """
        Checks meets_post_selection returns True for a valid State when multiple
        possible photon numbers are provided.
        """
        assert meets_post_selection(State([1, 0, 1, 0]), [((1,), (0, 1))])
        assert meets_post_selection(State([1, 1, 1, 0]), [((1,), (0, 1))])
        assert meets_post_selection(State([1, 1, 1, 0]), [((0, 1), (1, 2))])

    def test_meets_post_selection_multi_rules(self):
        """
        Checks meets_post_selection returns True for a valid State when multiple
        rules are all applied.
        """
        post_selection = [((1,), (0, 1)), ((2, 3), (1,))]
        assert meets_post_selection(State([1, 0, 1, 0]), post_selection)

    def test_meets_post_selection_multi_rules_false(self):
        """
        Checks meets_post_selection returns False for an invalid State when
        multiple rules are all applied.
        """
        post_selection = [((1,), (0, 1)), ((2, 3), (2,))]
        assert not meets_post_selection(State([1, 0, 1, 0]), post_selection)

    def test_meets_post_selection_empty_list(self):
        """
        Checks that meets_post_selection works with an empty list.
        """
        meets_post_selection(self.data, [])

    def test_apply_post_selection(self):
        """
        Checks apply_post_selection removes the correct States.
        """
        data = apply_post_selection(self.data["results"], [((0, 1), (1,))], [])
        for s in data:
            assert s[0] + s[1] == 1

    def test_apply_post_selection_multi_photon(self):
        """
        Checks apply_post_selection removes the correct States when multi photon
        numbers are valid.
        """
        data = apply_post_selection(
            self.data["results"], [((0, 1), (1, 2))], []
        )
        for s in data:
            assert s[0] + s[1] in {1, 2}

    def test_apply_post_selection_multi_rules(self):
        """
        Checks apply_post_selection removes the correct States when multi photon
        numbers are valid.
        """
        post_selection = [((0, 1), (1, 2)), ((2,), (1,))]
        data = apply_post_selection(self.data["results"], post_selection, [])
        for s in data:
            assert s[0] + s[1] in {1, 2}
            assert s[2] == 1

    def test_herald_modes_removed(self):
        """
        Checks that the herald modes are removed when this is used.
        """
        heralds = [0, 3]
        data = apply_post_selection(self.data["results"], [], heralds)
        for s in self.data["results"]:
            assert s[1:3] + s[4:] in data

    def test_apply_post_selection_empty_list(self):
        """
        Checks that apply_post_selection works with an empty list.
        """
        apply_post_selection(self.data, [], [])



# Copyright 2025 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import webbrowser
from argparse import ArgumentParser


def open_docs() -> None:
    """
    Open docs in a web browsers.
    """
    webbrowser.open("https://aegiq.gitlab.io/general/artemis/artemis-manual/")
    print("Documentation opened in browser.")  # noqa: T201


def main() -> None:
    """
    Processes provided arguments and performs corresponding commands.
    """
    parser = ArgumentParser(prog="lightworks_remote")
    _, uargs = parser.parse_known_args()
    for a in uargs:
        if a == "docs":
            open_docs()
        else:
            msg = f"Unrecognised argument '{a}' passed."
            raise ValueError(msg)


if __name__ == "__main__":
    main()



# Copyright 2025 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from typing import Any

from .job_details import JobDetails
from .status import Status

ARTEMIS_JOB_DETAILS_MAP = {
    "jobId": "job_id",
    "queuePosition": "queue_position",
    "jobStatus": "job_status",
    "createdDate": "created_date",
    "updatedDate": "updated_date",
    "logs": "logs",
    "userExecutionNotes": "notes",
}


@dataclass
class ArtemisJobDetails(JobDetails):
    """Stores all job details returned by Artemis"""

    job_id: int
    queue_position: int | None
    job_status: Status
    created_date: str
    updated_date: str
    logs: list[dict[str, Any]]
    notes: str | None

    def __post_init__(self) -> None:
        """Maps the provided status into one of the available statuses."""
        status_mapping = {
            "Accepted": "Accepted",
            "Scheduled": "Scheduled",
            "Running": "Running",
            "Completed": "Completed",
            "Failed": "Failed",
            "Cancelled": "Cancelled",
            "TimedOut": "Timed Out",
        }
        self.job_status = Status(status_mapping[self.job_status])  # type: ignore[index]


def get_no_results_reason(details: ArtemisJobDetails) -> str:
    """
    Generates a reason why no results were generated by a job, which can be
    appended to an error message.
    """
    if details.job_status in {
        Status.ACCEPTED,
        Status.SCHEDULED,
        Status.RUNNING,
    }:
        return "as the job is not yet complete."
    if details.job_status is Status.FAILED:
        if details.notes:
            return (
                f"as execution of the job failed with reason:\n{details.notes}"
            )
        return "as execution of the job failed."
    if details.job_status is Status.CANCELLED:
        return "as the job was cancelled."
    if details.job_status is Status.TIMEDOUT:
        return "as the job timed-out (likely from a hardware issue)."
    # Note this should never happen but is included just in case.
    return "for an unknown reason."



# Copyright 2025 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any

import numpy as np
from lightworks import State
from multimethod import multimethod

from lightworks_remote.payloads import Payload


@multimethod
def serialize(value: Any) -> Any:
    """
    Defines the default serialization behaviour in which the value is just
    returned without modification.
    """
    return value


@serialize.register
def serialize_payload(payload: Payload) -> dict[str, Any]:
    """
    Performs serialization of a dictionary to ensure it is compatible with
    submission via HTTP protocol.
    """
    return serialize(payload.payload)


@serialize.register
def serialize_complex(value: complex) -> str:
    """
    Converts a complex number into a string.
    """
    return str(value)


@serialize.register
def serialize_state(state: State) -> list[int]:
    """
    Retrieves the list representation of a State.
    """
    return list(state.s)


@serialize.register
def serialize_numpy_array(array: np.ndarray[Any, Any]) -> list[Any]:
    """
    Converts a numpy array into a list, casting the values to strings if the
    data type is complex.
    """
    if "complex" in str(array.dtype):
        array = array.astype(str)
    return array.tolist()


@serialize.register
def serialize_list(value: list[Any]) -> list[Any]:
    """
    Performs serialization of all values within a list.
    """
    for i, v in enumerate(value):
        value[i] = serialize(v)
    return value


@serialize.register
def serialize_dict(value: dict[Any, Any]) -> dict[str, Any]:
    """
    Performs serialization of all values within a dictionary.
    """
    return {str(serialize(k)): serialize(v) for k, v in value.items()}



# Copyright 2025 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Lightworks Remote
=================

Extension to lightworks to support compilation and submission of jobs to a
remote backend, either for emulation or hardware execution.
"""

import warnings

from .__version import __version__
from .remote import (
    QPU,
    cancel_job,
    check_job_complete,
    get_job_details,
    get_job_status,
    get_qpu_details,
    get_results,
    list_all_jobs,
    list_qpus,
    list_scheduled_jobs,
    load_job_from_file,
    show_scheduled,
)
from .token import token
from .utils import (
    install_certificate,
    saved_job_exists,
    set_api_url,
)
from .utils.exceptions import *

__all__ = [
    "QPU",
    "cancel_job",
    "check_job_complete",
    "get_job_details",
    "get_job_status",
    "get_qpu_details",
    "get_results",
    "install_certificate",
    "list_all_jobs",
    "list_qpus",
    "list_scheduled_jobs",
    "load_job_from_file",
    "saved_job_exists",
    "set_api_url",
    "show_scheduled",
    "token",
]


def warning_on_one_line(
    message: Warning | str,
    category: type[Warning],
    filename: str,  # noqa: ARG001
    lineno: int,  # noqa: ARG001
    line: str | None = None,  # noqa: ARG001
) -> str:
    """
    Customs warnings message which overwrites some of the noise in the default
    message.
    """
    return f"{category.__name__}: {message}\n"


warnings.formatwarning = warning_on_one_line



# Copyright 2025 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any

import pandas as pd

from lightworks_remote.http import ArtemisHTTPInterface, JobDetails
from lightworks_remote.results import QPUSamplingResult
from lightworks_remote.utils import get_data_from_json

from .batch_job import BatchJob
from .job import Job


def cancel_job(job_id: int | str) -> None:
    """
    Cancels job with matching id, raising an error if a matching job is not
    found.
    """
    api = ArtemisHTTPInterface()
    return api.cancel_job(job_id)


def list_scheduled_jobs() -> list[int]:
    """
    List all jobs awaiting execution on hardware.
    """
    api = ArtemisHTTPInterface()
    return api.list_scheduled_job_ids()


def list_all_jobs() -> list[int]:
    """
    List all jobs submitted by a user.
    """
    api = ArtemisHTTPInterface()
    return api.list_all_job_ids()


def get_job_status(job_id: int | str) -> str:
    """
    Gets status of job with a matching id.
    """
    api = ArtemisHTTPInterface()
    return api.get_job_status(job_id).value


def get_job_details(job_id: int | str) -> JobDetails:
    """
    Returns all details about the requested job.
    """
    api = ArtemisHTTPInterface()
    return api.get_job_details(job_id)


def check_job_complete(job_id: int | str) -> bool:
    """
    Returns whether or not the requested job is completed.
    """
    api = ArtemisHTTPInterface()
    return api.get_job_status(job_id).is_complete


def get_results(job_id: int | str) -> QPUSamplingResult:
    """
    Retrieves results for the job with matching id.
    """
    api = ArtemisHTTPInterface()
    return api.get_results(job_id)


def show_scheduled() -> pd.DataFrame:
    """
    Creates and returns a dataframe detailing all scheduled jobs and their
    current status.
    """
    # NOTE: Maybe this should only show a maximum number of jobs
    api = ArtemisHTTPInterface()
    scheduled = list_scheduled_jobs()
    all_data = []
    for s in scheduled:
        job_data = api.get_job_details(s)
        # If not scheduled yet then set position to be largest possible value
        position = job_data.queue_position
        position = position if position is not None else 2**31 - 1
        all_data.append([job_data.job_id, job_data.job_status.value, position])
    # Create dataframe and sort
    sch = pd.DataFrame(all_data, columns=["Job ID", "Status", "Queue Position"])
    sch = sch.sort_values(["Queue Position", "Job ID"], ascending=[True, True])
    # Convert queue position into int then str and convert any original none
    # values to waiting
    sch = sch.astype({"Queue Position": int})
    sch = sch.astype({"Queue Position": str})

    return sch.replace({"Queue Position": {str(2**31 - 1): "Waiting"}})


def load_job_from_file(filename: str) -> Job | BatchJob:
    """
    Loads job data from the provided file path.

    Args:

        filename (str) : The name of the data file to load, this can also
            contain a path to load from within a specific directory.

    """
    data = get_data_from_json(filename)
    if "job_type" not in data:
        raise KeyError(
            "Unable to identify job type from file data. json file should "
            "contain the key 'type'."
        )
    job: Any
    if data["job_type"] == "job":
        job = Job()
    elif data["job_type"] == "batch_job":
        job = BatchJob()
    else:
        raise ValueError("Job type from data file not recognised.")
    job.load_from_file(filename)
    return job



# Copyright 2025 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Will interact with Artemis system through HTTP protocol
"""

import time
from typing import Any

from lightworks_remote.http import ArtemisHTTPInterface, JobDetails, Status
from lightworks_remote.payloads import Payload
from lightworks_remote.results import QPUSamplingResult
from lightworks_remote.utils import (
    JobError,
    JobSubmissionError,
    get_data_from_json,
    saved_job_exists,
    write_data_to_json,
)
from lightworks_remote.utils.serialization import serialize


class Job:
    """
    Stores details of a submitted job, can be used to recover status info and
    then results once computation is complete.

    Args:

        payload (dict) : The payload which defines the required configuration
            for the job.

    """

    def __init__(self, payload: Payload | None = None) -> None:
        self._payload = payload
        self._interface = ArtemisHTTPInterface()

    def __str__(self) -> str:
        id_ = self._job_id if hasattr(self, "_job_id") else None
        return f"Job(id = {id_})"

    def __repr__(self) -> str:
        return "lightworks_remote." + str(self)

    @property
    def job_id(self) -> int:
        """Returns the id of the job."""
        if not hasattr(self, "_job_id"):
            return None  # type: ignore[return-value]
        return self._job_id

    @property
    def status(self) -> Status:
        """Returns the status of the job."""
        return self._interface.get_job_status(self.job_id)

    @property
    def complete(self) -> bool:
        """
        Returns whether or not a job has completed. This does not mean it was
        successful however, and could have failed or been cancelled.
        """
        return self.status.is_complete

    @property
    def success(self) -> bool:
        """
        Confirms a job was completed successfully.
        """
        return self.status == Status.COMPLETED

    @property
    def queue_position(self) -> int | None:
        """
        Returns the queue position of the job if it has one, otherwise returns
        None.
        """
        return self._interface.get_job_queue_position(self.job_id)

    @property
    def payload(self) -> Payload | None:
        """Returns the payload used to create the job."""
        return self._payload

    @property
    def details(self) -> JobDetails:
        """Returns the id of the job."""
        return self._interface.get_job_details(self.job_id)

    def _submit(self) -> None:
        """
        Submits the configured job to the web interface. This will re-submit
        the job and replace the id if this has already been submitted.
        """
        if self.payload is None:
            raise JobSubmissionError("No payload provided to job.")
        serialized = serialize(self.payload)
        self._job_id = self._interface.submit_job(serialized)

    def cancel(self) -> None:
        """Cancels the current job."""
        return self._interface.cancel_job(self.job_id)

    def retry(self) -> None:
        """Resubmits the job."""
        return self._submit()

    def wait_until_complete(self) -> None:
        """Pauses Python execution until the job is complete."""
        while not self.complete:
            time.sleep(0.5)

    def get_result(self) -> QPUSamplingResult:
        """
        Gets the result of a job..
        """
        return self._interface.get_results(self.job_id)

    def save_to_file(
        self, filename: str, allow_overwrite: bool = False
    ) -> None:
        """
        Saves the job id and payload to a file so that this can be restored at a
        later date.

        Args:

            filename (str) : The name to use for saving the job data, this can
                also contain a path to save it within a specific directory.

            allow_overwrite (bool) : Controls whether an exists saved job while
                be overwritten. Defaults to False.

        """
        if saved_job_exists(filename) and not allow_overwrite:
            raise ValueError(
                "A saved job with this name already exists. Either use a "
                "different name or set allow_overwrite to True to allow "
                "replacement."
            )
        write_data_to_json(self._get_job_data(), filename)

    def load_from_file(self, filename: str) -> None:
        """
        Loads job data from the provided file path.

        Args:

            filename (str) : The name of the data file to load, this can also
                contain a path to load from within a specific directory.

        """
        data = get_data_from_json(filename)

        try:
            self._job_id = data["job_id"]
            payload = data["payload"]
        except KeyError as e:
            raise JobError(
                "Unable to locate required job data within file."
            ) from e
        self._payload = Payload(**payload) if payload else None

    def _get_job_data(self) -> dict[str, Any]:
        """
        Returns a dictionary containing all data related to a job.
        """
        return {
            "job_type": "job",
            "job_id": self.job_id,
            "payload": self.payload,
        }

    def _check_successful_compilation(self) -> None:
        """
        Waits until compilation has completed and then checks for success.
        """
        # Wait until compiled
        while self.status in {Status.ACCEPTED, Status.COMPILING}:
            pass
        # Then check for any errors
        failed, errors = self._interface.check_for_compilation_errors(
            self.job_id
        )
        if failed:
            error_msg = (
                "The following errors occurred during the compilation process "
                f"for the job {self.job_id}: \n"
            )
            error_msg += "\tCode \tMessage \n"
            for e in errors:
                error_msg += f"\t{e[0]} \t{e[1]} \n"
            raise JobSubmissionError(error_msg)



# Copyright 2025 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
For management of certificates required to communicate with the API.
"""

import shutil
import warnings

from .exceptions import LightworksRemoteError
from .file_management import get_persistent_file_data_location


def set_api_url(url: str) -> None:
    """
    Saves the api url to a text file and then also assigns to the URL class.
    """
    URL._set(url)


def install_certificate(filename: str) -> None:
    """
    Installs the web certificate required for communication with the API to the
    correct location.
    """
    path = get_persistent_file_data_location()
    shutil.copy(filename, path / "certificate.pem")


class APIURL:
    """
    Stores and manages URL information, attempting to retrieve it on class
    initialisation.
    """

    def __init__(self) -> None:
        path = get_persistent_file_data_location()
        self._url: str | None
        try:
            with open(path / "api_url.txt", encoding="utf-8") as f:
                self._url = f.read()
        except Exception:  # noqa: BLE001
            warnings.warn(
                "Could not retrieve API URL, configure this with the "
                "'set_api_url' function.",
                stacklevel=1,
                category=UserWarning,
            )
            self._url = None

    @property
    def url(self) -> str:
        """Returns the currently set url."""
        if self._url is None:
            raise LightworksRemoteError(
                "API URL not set, this can be configured with the 'set_api_url'"
                "function."
            )
        return self._url

    def _set(self, url: str) -> None:
        path = get_persistent_file_data_location()
        with open(path / "api_url.txt", "w", encoding="utf-8") as f:
            f.write(url)
        self._url = url


# Configure URL class
URL = APIURL()



# Copyright 2025 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Configures a test certificate and api url so that unit tests can be run
successfully.
"""

from lightworks_remote import install_certificate, set_api_url

with open("test_cert.pem", "w", encoding="utf-8") as file:
    file.write("This is a certificate.")

install_certificate("test_cert.pem")

set_api_url("https://test/")



# Copyright 2025 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# ruff: noqa: PLW1514

import json
from pathlib import Path

from lightworks_remote.utils import (
    AuthenticationError,
    TokenError,
    get_persistent_file_data_location,
)


class Token:
    """
    Stores and provides token data, as well as utilities for saving and loading
    this data to a persistent location.
    """

    def __init__(self) -> None:
        self.__token: str | None = None

    def __str__(self) -> str:
        return str(self.__token)

    def __repr__(self) -> str:
        return f"lightworks_remote.token('{self.__token}')"

    def set(self, token: str) -> None:
        """
        Sets the current token. This should be provided as a string.
        """
        if not isinstance(token, str):
            raise TypeError("Token should be provided in string format.")
        self.__token = token

    def get(self) -> str:
        """
        Retrieves the currently set token value.
        """
        if self.__token is None:
            raise AuthenticationError(
                "Token has not been configured, this can be achieved with the "
                "set method (lightworks_remote.token.set('token'))."
            )
        return self.__token

    def save(self, name: str, overwrite: bool = False) -> None:
        """
        Save the token that is currently set, using a name to identify it which
        can be used in the load method later.

        Args:

            name (str) : The name which should be used to save the token.

            overwrite (bool, optional) : Dictates whether the token name can be
                used to overwrite an existing token. Defaults to False.

        Raises:

            TokenError : If an existing token if found with the set name and
                overwrite is set to False.

        """
        if not isinstance(name, str):
            raise TypeError("Save name should be a string")
        path = self._get_file_loc()
        # If file exists then load
        if path.exists():
            with open(path) as f:
                data = json.load(f)
        else:
            data = {}
        # Check for existing key if not allowed to overwrite
        if not overwrite and name in data:
            raise TokenError(
                "Existing token with chosen name found, select a different"
                " name or pass overwrite = True to method call."
            )
        # Then added to dictionary
        data[name] = self.get()
        # And update save file
        with open(path, "w") as f:
            json.dump(data, f, indent=4)

    def load(self, name: str) -> None:
        """
        Loads a saved token with provided name.

        Args:

            name (str) : The name of the token to load.

        Raises:

            TokenError : When a token matching the provided name is not found.

        """
        # Check file and token name exists
        self._check_save_loc_exists()
        path = self._get_file_loc()
        with open(path) as f:
            data = json.load(f)
        if name not in data:
            raise TokenError("Token with requested name not found.")
        # Update set token value
        self.set(data[name])

    def delete_saved(self, name: str) -> None:
        """
        Delete saved token with provided name.

        Args:

            name (str) : The name of the token to delete.

        Raises:

            TokenError : When a token matching the provided name is not found.

        """
        # Check file and token name exists
        self._check_save_loc_exists()
        path = self._get_file_loc()
        with open(path) as f:
            data = json.load(f)
        if name not in data:
            raise TokenError("Token with requested name not found.")
        # Remove token
        del data[name]
        # Update save file
        with open(path, "w") as f:
            json.dump(data, f, indent=4)

    def list_saved(self) -> None:
        """
        Prints a list of all currently saved tokens.
        """
        # Check file exists
        self._check_save_loc_exists()
        path = self._get_file_loc()
        # Get all token data and print
        with open(path) as f:
            data = json.load(f)
        print(list(data.keys()))  # noqa: T201

    def _check_save_loc_exists(self) -> None:
        """
        Checks if the token save location has been created.
        """
        path = self._get_file_loc()
        if not path.exists():
            raise TokenError("No tokens have been saved yet.")

    def _get_file_loc(self) -> Path:
        """
        Returns file file location for json used to store tokens.
        """
        # Set path according to platform
        path = get_persistent_file_data_location()
        return path / "tokens.json"
